# -*- coding: utf-8 -*-
from django.contrib import messages
from django.contrib.admin.utils import model_ngettext
from django.core.exceptions import FieldDoesNotExist, PermissionDenied
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class UserAdminMixin:
    def save_model(self, request, obj, form, change):
        if hasattr(obj, 'creator_id') and not change:
            obj.creator_id = request.user.id
        if hasattr(obj, 'last_modifier_id'):
            obj.last_modifier_id = request.user.id
        super().save_model(request, obj, form, change)

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)
        for instance in instances:
            if hasattr(instance, 'creator_id') and instance.creator_id is None:
                instance.creator_id = request.user.id
            if hasattr(instance, 'last_modifier_id'):
                instance.last_modifier_id = request.user.id
        super().save_formset(request, form, formset, change)


class StatusActionMixin:
    """
    Description: adds two admin actions: one to activate items by setting status=1, and the other to deactivate them.

    Usage: add this mixin to the inherited classes of the desired ModelAdmin class.

    Requirement: the Model linked to the ModelAdmin must have the 'status' attribute.
    """
    def activate_status(self, model_admin, request, queryset):
        if not self.has_change_permission(request):
            raise PermissionDenied
        params = {"status": 1}
        try:
            self.model._meta.get_field("date_last_modify")
            params["date_last_modify"] = timezone.now()
        except FieldDoesNotExist:
            pass
        try:
            self.model._meta.get_field("last_modifier")
            params["last_modifier"] = request.user
        except FieldDoesNotExist:
            pass
        try:
            updated = queryset.update(**params)
            self.message_user(request, _(
                "Successfully activated %(count)d %(items)s. Counting also the ones that were already active."
            ) % {
                "count": updated, "items": model_ngettext(self.opts, updated)
            }, messages.SUCCESS)
        except Exception as e:
            self.message_user(request, e, messages.ERROR)

    def deactivate_status(self, model_admin, request, queryset):
        if not self.has_change_permission(request):
            raise PermissionDenied
        params = {"status": 0}
        try:
            self.model._meta.get_field("date_last_modify")
            params["date_last_modify"] = timezone.now()
        except FieldDoesNotExist:
            pass
        try:
            self.model._meta.get_field("last_modifier")
            params["last_modifier"] = request.user
        except FieldDoesNotExist:
            pass
        try:
            updated = queryset.update(**params)
            self.message_user(request, _(
                "Successfully deactivated %(count)d %(items)s. Counting also the ones that were already inactive."
            ) % {
                "count": updated, "items": model_ngettext(self.opts, updated)
            }, messages.SUCCESS)
        except Exception as e:
            self.message_user(request, e, messages.ERROR)

    def get_actions(self, request):
        """
        Adds the action to the list of available actions.
        """
        actions = super().get_actions(request)
        if self.has_change_permission(request):
            actions.update(
                activate_status=(
                    self.activate_status,
                    "activate_status",
                    _("Activate selected {}").format(self.opts.verbose_name_plural),
                ),
                deactivate_status=(
                    self.deactivate_status,
                    "deactivate_status",
                    _("Deactivate selected {}").format(self.opts.verbose_name_plural),
                )
            )
        return actions


class ObfuscateAdminMixin:

    obfuscate_fields = None

    def obfuscate(self, model_admin, request, queryset):
        try:
            self.model.obfuscate_qs(queryset=queryset, pk=None, field_name_list=self.obfuscate_fields)
            self.message_user(
                request,
                _('Records obfuscated successfully.'),
                messages.SUCCESS
            )
        except Exception:
            self.message_user(
                request,
                _('Error message: it is not possible to obfuscate records'),
                messages.ERROR
            )
    obfuscate.short_description = _('Obfuscate records')

    def get_actions(self, request):
        """
        Adds the obfuscate action to the list of available actions.
        """
        actions = super().get_actions(request)
        actions.update(
            obfuscate=(
                self.obfuscate,
                "obfuscate",
                self.obfuscate.short_description
            )
        )
        return actions
