# -*- coding: utf-8 -*-

"""
DevEmailBackend
===============

Questo Backend permette di intercettare tutte le email inviate con send_email
e sovrascrivere il valore di alcuni campi applicando un set di configurazioni.

Quando è utile
----------------

in ambienti di staging per evitare di mandare email al cliente finale dirottandole ad un indirizzo di test


Esempio di utilizzo
-------------------

::

  EMAIL_SUBJECT_PREFIX = '[DEV:OFFICE2020.PROJECT.IT]' # Già presente nei vari progetti
  EMAIL_BACKEND = 'thx.core.mail.backends.DevEmailBackend'

  THX_CORE_DEV_EMAIL_RECIPIENT = ['code@thux.it', ]
  THX_CORE_DEV_EMAIL_BCC = ['code@thux.it', ]
  THX_CORE_DEV_EMAIL_CC = ['code@thux.it', ]

Classi
========

.. autoclass:: DevEmailBackend
   :members:

"""
import warnings

from django.conf import settings
from django.core.mail.backends.smtp import EmailBackend

from thx.core.deprecation import RemovedInThxVersion030Warning


class DevEmailBackend(EmailBackend):
    """
    Questo Backend permette di intercettare tutte le email inviate con send_email
    e sovrascrivere il valore di alcuni campi con la seguente configurazione:

    * email_message.to viene sostituito con settings.THX_CORE_DEV_EMAIL_RECIPIENT
    * email_message.bcc viene sostituito con settings.THX_CORE_DEV_EMAIL_BCC
    * email_message.cc viene sostituito con settings.THX_CORE_DEV_EMAIL_CC
    * email_message.subject viene preceduto dal settings settings.EMAIL_SUBJECT_PREFIX
    """
    def _send(self, email_message):  # pragma: no cover
        to = self.get_settings_value('THX_CORE_DEV_EMAIL_RECIPIENT', 'DEV_EMAIL_RECIPIENT')
        bcc = self.get_settings_value('THX_CORE_DEV_EMAIL_BCC', 'DEV_EMAIL_BCC')
        cc = self.get_settings_value('THX_CORE_DEV_EMAIL_CC', 'DEV_EMAIL_CC')

        email_message.to = to
        if email_message.bcc:
            email_message.bcc = bcc
        if email_message.cc:
            email_message.cc = cc
        email_message.subject = f'{settings.EMAIL_SUBJECT_PREFIX} {email_message.subject}'
        super()._send(email_message)

    @staticmethod
    def _warn_deprecated_setting(old_setting: str, new_setting: str):  # pragma: no cover
        warnings.warn(
            f"The settings {old_setting} is deprecated, please use {new_setting}",
            category=RemovedInThxVersion030Warning,
            stacklevel=2,
        )

    def get_settings_value(self, new_setting: str, deprecated_setting: str):
        value = getattr(settings, deprecated_setting, None)
        if value:
            self._warn_deprecated_setting(deprecated_setting, new_setting)
        else:
            value = getattr(settings, new_setting, None)
        return value
