# -*- coding: utf-8 -*-
import logging

from django.conf import settings
from django.core.management.base import BaseCommand


class ThxBaseCommand(BaseCommand):
    """ThxBaseCommand

    Provides default logging behaviour to be used in all custom commands.

    The default logger is `file` and is token from settings `THX_CORE_COMMANDS_LOGGER`.
    You can change your logger overriding the setting or simply overriding the `logger` attribute.
    This class provides the default parsed arguments:
        -c --console: write logging to console
        -l --debug-level: set the logging level for console (default: `info`).
    """

    logger = logging.getLogger(settings.THX_CORE_DEFAULT_COMMANDS_LOGGER)

    def execute(self, *args, **options):
        self.set_logging(options)
        return super().execute(*args, **options)

    def create_parser(self, prog_name, subcommand, **kwargs):
        parser = super().create_parser(prog_name, subcommand, **kwargs)
        parser.add_argument(
            '-c', '--console', action='store_true', default=False,
            help='Debug - write logging to console',
        )
        parser.add_argument(
            '-l', '--debug-level', default='info',
            help='Set debug level (debug, info, warnings) for console',
        )
        return parser

    def set_logging(self, options: dict) -> None:
        debug_level = getattr(logging, options.get('debug_level').upper())
        self.logger.handlers[0].setLevel(debug_level)

        if options.get('console'):
            console_handler = logging.StreamHandler()
            console_handler.setLevel(debug_level)
            self.logger.addHandler(console_handler)
        self.logger.setLevel(debug_level)
