# -*- coding: utf-8 -*-

'''
Db Mixins
============

Per semplificare la creazione dei nostri modelli e per renderla più omogenea sono
stati creati dei mixin astratti (``Meta:  abstract=True``) che aggiungono campi e
funzionalità (vedi in particolare :class:`CleanModel`):

* :class:`UserModel` (``creator``, ``last_modifier``)
* :class:`DateModel` (``date_create``, ``date_last_modify``)
* :class:`StatusModel` (``status``)
* :class:`OrderedModel` (``ordering``)
* :class:`SEOModel` (``meta_description``, ``page_title``, ``meta_keywords``)
* :class:`CleanModel` (sovrascive ``.save()``)
* :class:`DeleteFileModel` (sovrascive ``.save()`` e ``.delete()``)
* :class:`ObfuscateModel` (``.obfuscate()`` e ``.obfuscate_qs()``)

Esempio
--------

::

  class PriceList(CleanModel, UserModel, DateModel, StatusModel, OrderedModel):
      """
      Price List belongs to Stores, not to Suppliers. A supplier can have different PriceList for
      different Stores
      """
      COST_PRICE = settings.PRICE_PRICE_LIST_COST_PRICE
      RESERVED_PRICE = settings.PRICE_PRICE_LIST_RESERVED_PRICE
      STREET_PRICE = settings.PRICE_PRICE_LIST_STREET_PRICE

      objects = models.Manager()
      available = managers.AvailablePriceListManager()

      code = models.CharField(
          max_length=100,
          verbose_name=_('code'),
      )


Modelli
========

.. autoclass:: UserModel
   :members:

.. autoclass:: DateModel
   :members:

.. autoclass:: StatusModel
   :members:

.. autoclass:: OrderedModel
   :members:

.. autoclass:: SEOModel
   :members:

.. autoclass:: CleanModel
   :members:

.. autoclass:: DeleteFileModel
   :members:

.. autoclass:: ObfuscateModel
   :members:

'''
import os
import warnings

from ckeditor.fields import RichTextField
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import connection, models, transaction
from django.utils.translation import gettext_lazy as _

from thx.core.deprecation import RemovedInThxVersion030Warning


class UserModel(models.Model):
    """
    **UserModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede vengono aggiunti due attributi entrambi di tipo settings.AUTH_USER_MODEL:

    :creator: indica l'utente che ha creato il record; una volta inserito non dovrà più essere aggiornato;
    :last_modifier: indica l'ultimo utente in ordine temporale che ha modificato il record

    Entrambe le FK sono definite on_delete=models.PROTECT
    """

    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="%(app_label)s_%(class)s_creator",
        verbose_name=_("creator"),
    )
    last_modifier = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name='%(app_label)s_%(class)s_last_modifier',
        verbose_name=_("last modifier"),
    )

    class Meta:
        abstract = True


class DateModel(models.Model):
    """
    **DateModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede vengono aggiunti due attributi entrambi di tipo settings.DateTimeField:

    :date_create: indica il datetime di creazione; viene valorizzato in modo automatico durante la creazione del record
        non verrà più aggiornato

    :date_last_modify: indica il datetime di ultima modifica; viene valorizzato ad ogni salvataggio con il datetime
        corrente

    """

    date_create = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('date create'),
    )
    date_last_modify = models.DateTimeField(
        auto_now=True,
        verbose_name=_('date last modify'),
    )

    class Meta:
        abstract = True


class StatusModel(models.Model):
    """
    **StatusModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede viene aggiunto un attributo:

    - status: indica se un record è enable o disable e i valori sono limitati dal settings STATUS_CODES;
        viene utilizzato per la cancellazione logica di un record
    """

    STATUS_CODES = (
        (0, _('disable')),
        (1, _('enable')),
    )
    status = models.IntegerField(
        choices=STATUS_CODES,
        default=1,
        verbose_name=_('status'),
        null=True
    )

    class Meta:
        abstract = True

    @property
    def status_display(self) -> str:
        return self.get_status_display()

    def status_icon(self):
        return self.status
    status_icon.short_description = _('status')
    status_icon.admin_order_field = 'status'
    status_icon.boolean = True


class OrderedModel(models.Model):
    """
    **OrderedModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede viene aggiunto un attributo:

    - ordering: serve per dare una posizione fissa al record nell'ordine del queryset
    """

    ordering = models.IntegerField(
        default=0,
        verbose_name=_('ordering'),
    )

    class Meta:
        abstract = True


class SEOModel(models.Model):
    """
    **SEOModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede vengono aggiunti tre attributi che possono essere utilizzati per il SEO
    """

    meta_description = models.TextField(
        _("description"),
        max_length=255,
        blank=True,
    )
    meta_keywords = models.CharField(
        _("keywords"),
        max_length=255,
        blank=True,
    )
    page_title = models.CharField(
        _("title"),
        max_length=255,
        help_text=_("overwrite the title (html title tag)"),
        blank=True,
    )

    class Meta:
        abstract = True


class CleanModel(models.Model):
    """
    **CleanModel** è una classe *abstract* che estende django.db.models.Model.

    viene ridefinita la funzione 'save' nella quale viene richiamata la full_clean (chiama tutte le clean() in cascata)
    per validare gli attributi del record prima della 'save' effettiva
    """

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)


class MaterializedDbViewModel(models.Model):
    """Docs: https://www.postgresql.org/docs/13/rules-materializedviews.html"""

    # TODO valutare se deprecare in favore di https://github.com/muehlemann-popp/django-materialized-view.
    #  Per ora non testare

    @classmethod
    def get_sql_create(cls, with_data: bool = True):
        return "CREATE MATERIALIZED VIEW {} AS {} {}".format(
            cls._meta.db_table,
            str(cls.get_materialized_view_queryset().query),
            "WITH DATA" if with_data else "WITH NO DATA"
        )

    @classmethod
    def get_sql_delete(cls):
        return "DROP MATERIALIZED VIEW IF EXISTS {}".format(cls._meta.db_table)

    @classmethod
    def refresh(cls, concurrently: bool = False):
        """
        From: https://www.postgresql.org/docs/13/sql-refreshmaterializedview.html

        :param concurrently: Refresh the materialized view without locking out concurrent selects on the materialized
         view. Without this option a refresh which affects a lot of rows will tend to use fewer resources and complete
          more quickly, but could block other connections which are trying to read from the materialized view.
          This option may be faster in cases where a small number of rows are affected.
          This option is only allowed if there is at least one UNIQUE index on the materialized view which uses
          only column names and includes all rows; that is, it must not be an expression index or
          include a WHERE clause.
          This option may not be used when the materialized view is not already populated.
          Even with this option only one REFRESH at a time may run against any one materialized view.
        """
        cursor_wrapper = connection.cursor()
        cursor = cursor_wrapper.cursor
        try:
            if concurrently:
                cursor.execute("REFRESH MATERIALIZED VIEW CONCURRENTLY {}".format(cls._meta.db_table))
            else:
                cursor.execute("REFRESH MATERIALIZED VIEW {}".format(cls._meta.db_table))
        finally:
            cursor_wrapper.close()

    @classmethod
    def get_materialized_view_queryset(cls):
        raise NotImplementedError("`get_materialized_view_queryset` must be overwritten")

    class Meta:
        abstract = True
        managed = False


class DeleteFileModel(models.Model):
    """
    **DeleteFileModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede vengono ridefiniti i metodi:
    - delete: oltre alla cancellazione dell'istanza, elimina i files collegati ai FileField del modello
    - save: elimina i files collegati ai FileField del modello qualora questi vengano aggiornati

    Alla classe erede vengono aggiunti i metodi:
    - delete_files: elimina i files contenuti nell'iterabile ricevuto in input
    - delete_file_on_upload: elimina i files contenuti nell'iterabile ricevuto in input
    """

    class Meta:
        abstract = True

    @transaction.atomic
    def delete(self, *args, **kwargs):
        path_list = []
        if settings.THX_CORE_DELETE_FILE_MODEL_DELETE_FILE_ON_DELETE:
            # TODO: da valutare se all'interno del settings è possibile inserire solo i campi da escludere
            for key, value in settings.THX_CORE_DELETE_FILE_MODEL_DELETE_FILE_ON_DELETE_FIELDS.items():
                if self._meta.model_name == key:
                    for field_name in value:
                        field = self._meta.get_field(field_name)
                        if isinstance(field, models.FileField):
                            path_list.append(getattr(self, field.name).path)

        super().delete(*args, **kwargs)

        if settings.THX_CORE_DELETE_FILE_MODEL_DELETE_FILE_ON_DELETE:
            self.delete_files(path_list)

    @staticmethod
    def delete_files(path_list):
        for path in path_list:
            if os.path.isfile(path):
                os.remove(path)

    def save(self, *args, **kwargs):
        if self.pk:
            if settings.THX_CORE_DELETE_FILE_MODEL_DELETE_FILE_ON_UPLOAD:
                for field in self._meta.get_fields():
                    if isinstance(self._meta.get_field(field.name), models.FileField):
                        obj = self._meta.model.objects.get(pk=self.pk)
                        if getattr(obj, field.name) is not None:
                            self.delete_file_on_upload(field.name)

        super().save(*args, **kwargs)

    def delete_file_on_upload(self, attribute_name):
        try:
            old_instance = self._meta.model.objects.get(pk=self.pk)
            try:
                file = getattr(old_instance, attribute_name)
                if file:
                    file.delete(save=True)
            except:  # todo gestire eccezioni
                pass
        except self._meta.model.DoesNotExist:
            pass


class ObfuscateModel(models.Model):
    """
    **ObfuscateModel** è una classe *abstract* che estende django.db.models.Model.

    Alla classe erede vengono aggiunti i metodi:
    - obfuscate: offusca i campi dell'istanza
    - obfuscate_qs: (class method) offusca i campi del queryset
    """
    class Meta:
        abstract = True

    def obfuscate(self, field_name_list=None):
        """Obfuscate a single instance"""
        self.obfuscate_qs(pk=self.pk, field_name_list=field_name_list)

    @classmethod
    @transaction.atomic()
    def obfuscate_qs(cls, queryset=None, pk=None, field_name_list=None):
        """Obfuscate an entire queryset"""
        if queryset is None:
            if not pk:
                raise ValidationError("One between `queryset` and `pk` arguments should be provided.")
            queryset = cls._meta.model.objects.filter(pk=pk)
        else:
            if pk:
                raise ValidationError("Only one between `queryset` and `pk` arguments should be provided.")

        fields = (
            [cls._meta.get_field(field_name) for field_name in field_name_list] if field_name_list
            else cls._meta.get_fields()
        )

        for obj in queryset:
            for field in fields:
                if not getattr(field, 'primary_key', False):
                    if field.name == 'status':
                        setattr(obj, field.name, 0)
                    elif isinstance(field, models.IntegerField):
                        setattr(
                            obj, field.name, field.choices[0][0]
                        ) if field.choices else setattr(obj, field.name, 0)
                    elif isinstance(field, models.EmailField):
                        setattr(obj, field.name, 'deleted_{}@deleted.com'.format(obj.pk))
                    elif isinstance(field, models.CharField):
                        setattr(obj, field.name, 'deleted'[:field.max_length])
                    elif isinstance(field, models.BooleanField):
                        setattr(obj, field.name, False)
                    elif isinstance(field, models.DateField) or isinstance(field, models.DateTimeField):
                        if field.null:
                            setattr(obj, field.name, None)
                    elif isinstance(field, models.TextField) or isinstance(field, RichTextField):
                        setattr(obj, field.name, 'deleted')
            obj.save()


# DEPRECATIONS thx-core 0.3.0
class ObfuscateMixin(object):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "thx.core.models.ObfuscateMixin is deprecated in favour of thx.core.models.ObfuscateModel. "
            f"Please update model {self.__class__.__name__}",
            category=RemovedInThxVersion030Warning,
            stacklevel=2,
        )

    def obfuscate(self, request=None, queryset=None, pk=None):
        from django.contrib import messages
        try:
            ids = list(queryset.values_list('id', flat=True)) if queryset else [pk, ]
            for id in ids:
                record = self.model.objects.get(pk=id)
                for field in record._meta.get_fields():
                    if field.name == 'status':
                        setattr(record, field.name, 0)
                    if type(field) == models.IntegerField:
                        setattr(
                            record, field.name, field.choices[0][0]
                        ) if field.choices else setattr(record, field.name, 0)
                    elif type(field) == models.CharField:
                        setattr(record, field.name, 'deleted'[:field.max_length])
                    elif type(field) == models.EmailField:
                        setattr(record, field.name, 'deleted_{}@deleted.com'.format(record.pk))
                    elif type(field) == models.BooleanField:
                        setattr(record, field.name, False)
                    elif type(field) == models.DateField or type(field) == models.DateTimeField:
                        if field.null:
                            setattr(record, field.name, None)
                    elif type(field) == models.TextField or type(field) == RichTextField:
                        setattr(record, field.name, 'deleted')
                    else:
                        pass
            self.message_user(
                request,
                _('Records obfuscated successfully.'),
                messages.SUCCESS
            )
        except:
            self.message_user(
                request,
                _('Error message: it is not possible to obfuscate records'),
                messages.ERROR
            )

    obfuscate.short_description = _('Obfuscate records')


class DeleteFileMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "thx.core.models.DeleteFileMixin is deprecated in favour of thx.core.models.DeleteFileModel. "
            f"Please update model {self.__class__.__name__}",
            category=RemovedInThxVersion030Warning,
            stacklevel=2,
        )

    @transaction.atomic
    def delete(self, *args, **kwargs):
        path_list = []
        if settings.THX_CORE_DELETE_FILE_MIXIN_DELETE_FILE_ON_DELETE:
            # TODO: da valutare se all'interno del settings è possibile inserire solo i campi da escludere
            for key, value in settings.THX_CORE_DELETE_FILE_MIXIN_DELETE_FILE_ON_DELETE_FIELDS.items():
                if self._meta.model_name == key:
                    for field_name in value:
                        field = self._meta.get_field(field_name)
                        if isinstance(field, models.FileField):
                            path_list.append(getattr(self, field.name).path)

        super().delete(*args, **kwargs)

        if settings.THX_CORE_DELETE_FILE_MIXIN_DELETE_FILE_ON_DELETE:
            self.delete_files(path_list)

    @staticmethod
    def delete_files(path_list):
        for path in path_list:
            if os.path.isfile(path):
                os.remove(path)

    def save(self, *args, **kwargs):
        if self.pk:
            if settings.THX_CORE_DELETE_FILE_MIXIN_DELETE_FILE_ON_UPLOAD:
                for field in self._meta.get_fields():
                    if isinstance(self._meta.get_field(field.name), models.FileField):
                        obj = self._meta.model.objects.get(pk=self.pk)
                        if getattr(obj, field.name) is not None:
                            self.delete_file_on_upload(field.name)

        super().save(*args, **kwargs)

    def delete_file_on_upload(self, attribute_name):
        try:
            old_instance = self._meta.model.objects.get(pk=self.pk)
            try:
                file = getattr(old_instance, attribute_name)
                if file:
                    file.delete(save=True)
            except:  # todo gestire eccezioni
                pass
        except self._meta.model.DoesNotExist:
            pass
