# -*- coding: utf-8 -*-
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class JSONSchemaValidator:
    """JSONSchemaValidator

    Require `jsonschema` to be installed in project.

    Usage:

    INFO_SCHEMA = {
        "type": "object",
        "properties": {
            "field_1": {"type": "array"},
            "field_2": {"type": "boolean"},
        }
    }

    class MyModel:
        info = models.JSONField(
            validators=[JSONSchemaValidator(INFO_SCHEMA)],
            verbose_name=_("info"),
        )
    """

    message = _("%(value)s failed JSON schema check. JSON schema Exception error: %(exc)s")
    code = "invalid_json"

    def __init__(self, schema: dict) -> None:
        """
        JSONSchemaValidator constructor

        :param schema: JSON schema through which validate the JSON field data.
        """
        self.schema = schema
        self._check_jsonschema_installed()
        self._validate_schema()

    def __call__(self, value: dict) -> None:
        """
        Apply JSON Schema validation.

        :param value: JSON data to be validated through the JSON schema supplied during class initialization.
        :raise: django.core.exceptions.ValidationError in case of a jsonschema.exceptions.ValidationError error is
            raised.
        """
        import jsonschema

        try:
            jsonschema.validate(value, self.schema)
        except jsonschema.exceptions.ValidationError as exc:
            raise ValidationError(
                message=self.message,
                code=self.code,
                params={
                    "value": value,
                    "exc": exc,
                },
            )

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.schema == other.schema

    def _validate_schema(self) -> None:
        """Validate schema"""
        import jsonschema

        try:
            jsonschema.validate({}, self.schema)
        except jsonschema.exceptions.SchemaError as e:
            raise ImproperlyConfigured(f"JSON Schema {self.schema} is invalid. Exception:\n{str(e)}") from e
        except jsonschema.exceptions.ValidationError:
            pass

    @staticmethod
    def _check_jsonschema_installed():
        try:
            import jsonschema  # noqa
        except ImportError:  # pragma: no cover
            raise ImproperlyConfigured(
                "Cannot use JSONSchemaValidator because jsonschema is not installed."
                "Get jsonschema at https://pypi.org/project/jsonschema/ "
                'or run command "python -m pip install jsonschema".'
            )
